from typing import Any, Optional, Protocol

from carnival import global_context


class _Writer(Protocol):
    def write(self, __s: str) -> Any: ...


def log(message: str, file: Optional[_Writer] = None) -> None:
    if global_context.host is None:
        host = "NO CONNECTION"
    else:
        host = global_context.host.host

    print(f"💃💃💃 {host}> {message}", file=file)



from carnival.step import Step
from carnival.host import Host, SSHHost, LocalHost
from carnival.task import Task, SimpleTask
from carnival import cmd
from carnival import internal_tasks
from carnival.utils import log
from carnival.context import context_ref

__all__ = [
    'Step',
    'Host', 'SSHHost', 'LocalHost',
    'Task', 'SimpleTask',
    'cmd',
    'log',
    'context_ref',
    'internal_tasks',
]



from typing import List, Optional

from carnival import cmd, global_context
from invoke import Result  # type: ignore
from patchwork import files  # type:ignore


def mkdirs(*dirs: str) -> List[Result]:
    """
    Создать директории

    :param dirs: пути которые нужно создать
    """
    return [cmd.cli.run(f"mkdir -p {x}", hide=True) for x in dirs]


def is_dir_exists(dir_path: str) -> bool:
    """
    Узнать существует ли директория

    :param dir_path: путь до директории
    """
    return bool(cmd.cli.run(f"test -d {dir_path}", warn=True, hide=True).ok)


def is_file_contains(filename: str, text: str, exact: bool = False, escape: bool = True) -> bool:
    """
    Содержит ли файл текст
    См <https://fabric-patchwork.readthedocs.io/en/latest/api/files.html#patchwork.files.contains>

    :param filename: путь до файла
    :param text: текст который нужно искать
    :param exact: точное совпадение
    :param escape: экранировать ли текст


    """
    assert global_context.conn is not None, "No connection"
    return bool(files.contains(
        global_context.conn,
        runner=global_context.conn.run,
        filename=filename, text=text, exact=exact, escape=escape
    ))


def is_file_exists(path: str) -> bool:
    """
    Проверить существует ли файл
    <https://fabric-patchwork.readthedocs.io/en/latest/api/files.html#patchwork.files.exists>

    :param path: путь до файла
    """
    assert global_context.conn is not None, "No connection"
    return bool(files.exists(global_context.conn, runner=global_context.conn.run, path=path))


def ensure_dir_exists(
    path: str,
    user: Optional[str] = None,
    group: Optional[str] = None,
    mode: Optional[str] = None,
) -> None:
    """
    Проверить что директория существует и параметры соответствуют заданным

    <https://fabric-patchwork.readthedocs.io/en/latest/api/files.html#patchwork.files.directory>

    :param path: путь до директории
    :param user: владелец
    :param group: группа
    :param mode: права
    """
    assert global_context.conn is not None, "No connection"
    files.directory(global_context.conn, runner=global_context.conn.run, path=path, user=user, group=group, mode=mode)



"""
Объект, представляющий единицу оборудования.

Carnival не предоставляет никаких сложных абстракций для работы с группами хостов,
подразумевая что вы будете использовать встроенные коллекции python и организуете
работу так, как будет удобно для вашей задачи.
В простом случае, можно передавать хосты прямо в коде файла `carnival_tasks.py`.

>>> class SetupFrontend(Task):
>>>    def run(self, **kwargs):
>>>        self.step(Frontend(), SSHHost("1.2.3.4", packages=["htop", ]))

В более сложных, создать списки в файле `inventory.py`

>>> # inventory.py
>>> frontends = [
>>>     LocalHost(),
>>>     SSHHost("1.2.3.5"),
>>> ]

>>> # carnival_tasks.py
>>> import inventory as i
>>> class SetupFrontend(Task):
>>>    def run(self, **kwargs):
>>>        self.step(Frontend(), i.frontends)
"""

from typing import Any, Optional, Union
import warnings

from fabric.connection import Connection as SSHConnection  # type: ignore
from invoke.context import Context as LocalConnection  # type: ignore
from paramiko.client import MissingHostKeyPolicy, AutoAddPolicy  # type: ignore


AnyConnection = Union[SSHConnection, LocalConnection]

"""
Список адресов которые трактуются как локальное соединение
.. deprecated:: 1.4
        Host is deprecated, use LocalHost or SSHHost explicitly
"""
LOCAL_ADDRS = [
    'local',
    'localhost',
]


class LocalHost:
    """
    Локальный хост, работает по локальному терминалу
    """

    def __init__(self, **context: Any) -> None:
        self.addr = "local"
        self.context = context

    def connect(self) -> LocalConnection:
        return LocalConnection()

    @property
    def host(self) -> str:
        """
        Remove user and port parts, return just address
        """

        h = self.addr

        if ':' in self.addr:
            h = h.split(":", maxsplit=1)[0]

        return h

    def __str__(self) -> str:
        return f"🖥 {self.host}"

    def __hash__(self) -> int:
        return hash(self.addr)

    def __repr__(self) -> str:
        return f"<Host object {self.host}>"


class SSHHost(LocalHost):
    """
    Удаленный хост, работает по SSH
    """

    def __init__(
        self,
        addr: str,
        ssh_user: Optional[str] = None, ssh_password: Optional[str] = None, ssh_port: int = 22,
        ssh_gateway: Optional['SSHHost'] = None,
        ssh_connect_timeout: int = 10,
        missing_host_key_policy: MissingHostKeyPolicy = AutoAddPolicy,

        **context: Any
     ):
        """
        :param addr: Адрес сервера
        :param ssh_user: Пользователь SSH
        :param ssh_password: Пароль SSH
        :param ssh_port: SSH порт
        :param ssh_connect_timeout: SSH таймаут соединения
        :param ssh_gateway: Gateway
        :param context: Контекст хоста
        """
        if "@" in addr:
            raise ValueError("Please set user in 'ssh_user' arg")

        self.addr = addr
        self.ssh_port = ssh_port
        self.context = context
        self.ssh_user = ssh_user
        self.ssh_password = ssh_password
        self.ssh_connect_timeout = ssh_connect_timeout
        self.ssh_gateway: Optional['SSHHost'] = ssh_gateway
        self.missing_host_key_policy = missing_host_key_policy

    def is_connection_local(self) -> bool:
        """
        Check if host's connection is local
        """
        warnings.warn(
            "is_connection_local is deprecated, use LocalHost or SSHHost explicitly",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.host.lower() in LOCAL_ADDRS

    def connect(self) -> SSHConnection:
        gateway = None
        if self.ssh_gateway:
            gateway = self.ssh_gateway.connect()
            assert isinstance(gateway, SSHConnection), f"{self.ssh_gateway} is not ssh connection"

        conn = SSHConnection(
            host=self.addr,
            port=self.ssh_port,
            user=self.ssh_user,
            connect_timeout=self.ssh_connect_timeout,
            gateway=gateway,
            connect_kwargs={
                'password': self.ssh_password,
            }
        )
        conn.client.set_missing_host_key_policy(self.missing_host_key_policy)
        return conn


AnyHost = Union[LocalHost, SSHHost]


class Host(SSHHost):
    """
    :param addr: Адрес сервера для SSH или "local" для локального соединения
    :param ssh_user: Пользователь SSH
    :param ssh_password: Пароль SSH
    :param ssh_port: SSH порт
    :param ssh_connect_timeout: SSH таймаут соединения
    :param ssh_gateway: Gateway
    :param context: Контекст хоста

    .. deprecated:: 1.4
        Host is deprecated, use LocalHost or SSHHost explicitly
    """
    def __init__(
        self,
        addr: str,
        ssh_user: Optional[str] = None, ssh_password: Optional[str] = None, ssh_port: int = 22,
        ssh_gateway: Optional['SSHHost'] = None, ssh_connect_timeout: int = 10,
        missing_host_key_policy: MissingHostKeyPolicy = AutoAddPolicy,
        **context: Any
    ):
        warnings.warn(
            "Host is deprecated, use LocalHost or SSHHost explicitly",
            DeprecationWarning,
            stacklevel=2,
        )

        if "@" in addr:
            ssh_user, addr = addr.split("@", maxsplit=1)

        super().__init__(
            addr,
            ssh_user=ssh_user, ssh_password=ssh_password, ssh_port=ssh_port,
            ssh_gateway=ssh_gateway, ssh_connect_timeout=ssh_connect_timeout,
            missing_host_key_policy=missing_host_key_policy, **context
        )

    def connect(self) -> AnyConnection:
        if self.addr in LOCAL_ADDRS:
            return LocalConnection()

        return super().connect()



from carnival import Task


class Help(Task):
    """
    Показать список доступных задач
    """
    module_name = ""
    help = "List available commands and help"

    def run(self) -> None:
        from carnival.cli import task_types

        task_list = list(task_types.keys())
        task_list.sort()

        ralign = 4
        if task_list:
            ralign += max([len(x) for x in task_list])

        for task_name in task_types.keys():
            help_text = task_types[task_name].help
            if help_text:
                print(f" {task_name:<{ralign}} {help_text}")
            else:
                print(f" {task_name}")



from typing import List, Optional

from carnival import cmd
from carnival.utils import log


def get_pkg_versions(pkgname: str) -> List[str]:
    """
    Получить список доступных версий пакета
    """
    versions = []
    result = cmd.cli.run(f"DEBIAN_FRONTEND=noninteractive apt-cache madison {pkgname}", hide=True, warn=True)
    if result.ok is False:
        return []

    for line in result.stdout.strip().split("\n"):
        n, ver, r = line.split("|")
        versions.append(ver.strip())
    return versions


def get_installed_version(pkgname: str) -> Optional[str]:
    """
    Получить установленную версию пакета

    :return: Версия пакета если установлен, `None` если пакет не установлен
    """
    result = cmd.cli.run(f"DEBIAN_FRONTEND=noninteractive dpkg -l {pkgname}", hide=True, warn=True)
    if result.ok is False:
        return None
    installed, pkgn, ver, arch, *desc = result.stdout.strip().split("\n")[-1].split()
    if installed != 'ii':
        return None

    assert isinstance(ver, str)
    return ver.strip()


def is_pkg_installed(pkgname: str, version: Optional[str] = None) -> bool:
    """
    Проверить установлен ли пакет
    Если версия не указана - проверяется любая
    """

    pkgver = get_installed_version(pkgname)
    if version is None and pkgver is not None:
        return True

    if version is not None and pkgver == version:
        return True

    return False


def force_install(pkgname: str, version: Optional[str] = None, update: bool = False, hide: bool = False) -> None:
    """
    Установить пакет без проверки установлен ли он
    """
    if version:
        pkgname = f"{pkgname}={version}"

    if update:
        cmd.cli.run("DEBIAN_FRONTEND=noninteractive sudo apt-get update", pty=True, hide=hide)

    cmd.cli.run(f"DEBIAN_FRONTEND=noninteractive sudo apt-get install -y {pkgname}", pty=True, hide=hide)


def install(pkgname: str, version: Optional[str] = None, update: bool = True, hide: bool = False) -> bool:
    """
    Установить пакет если он еще не установлен в системе

    :param pkgname: название пакета
    :param version: версия
    :param update: запустить apt-get update перед установкой
    :param hide: скрыть вывод этапов
    :return: `True` если пакет был установлен, `False` если пакет уже был установлен ранее
    """
    if is_pkg_installed(pkgname, version):
        if version:
            if not hide:
                log(f"{pkgname}={version} already installed")
        else:
            if not hide:
                log(f"{pkgname} already installed")
        return False
    force_install(pkgname=pkgname, version=version, update=update, hide=hide)
    return True


def install_multiple(*pkg_names: str, update: bool = True, hide: bool = False) -> bool:
    """
    Установить несколько пакетов, если они не установлены

    :param pkg_names: список пакетов которые нужно установить
    :param update: запустить apt-get update перед установкой
    :param hide: скрыть вывод этапов
    :return: `True` если хотя бы один пакет был установлен, `False` если все пакеты уже были установлен ранее
    """
    if all([is_pkg_installed(x) for x in pkg_names]):
        if not hide:
            log(f"{','.join(pkg_names)} already installed")
        return False

    if update:
        cmd.cli.run("DEBIAN_FRONTEND=noninteractive sudo apt-get update", pty=True, hide=hide)

    for pkg in pkg_names:
        install(pkg, update=False, hide=hide)
    return True


def remove(*pkg_names: str, hide: bool = False) -> None:
    """
    Удалить пакет

    :param pkg_names: список пакетов которые нужно удалить
    :param hide: скрыть вывод этапов
    """
    assert pkg_names, "pkg_names is empty"
    cmd.cli.run(f"DEBIAN_FRONTEND=noninteractive sudo apt-get remove --auto-remove -y {' '.join(pkg_names)}", hide=hide)



import os
from typing import List

from carnival import cmd
from invoke import Result  # type: ignore


def set_password(username: str, password: str) -> Result:
    """
    Установить пароль пользователю

    :param username: Пользователь
    :param password: Новый пароль
    """
    return cmd.cli.pty(f"echo '{username}:{password}' | chpasswd", hide=True)


def ssh_authorized_keys_add(ssh_key: str, keys_file: str = ".ssh/authorized_keys") -> bool:
    """
    Добавить ssh ключ в `authorized_keys`

    :param ssh_key: ключ
    :param keys_file: пусть до файла `authorized_keys`
    :return: `True` если ключ был добавлен, `False` если ключ уже был в файле
    """
    ssh_key = ssh_key.strip()

    cmd.cli.run("mkdir -p ~/.ssh")
    cmd.cli.run("chmod 700 ~/.ssh")
    cmd.cli.run(f"touch {keys_file}")

    if not cmd.fs.is_file_contains(keys_file, ssh_key, escape=True):
        cmd.cli.run(f"echo '{ssh_key}' >> {keys_file}")
        return True
    return False


def ssh_authorized_keys_list() -> List[str]:
    """
    Получить список авторизованных ssh-ключей сервера
    """
    if cmd.fs.is_file_exists("~/.ssh/authorized_keys") is False:
        return []

    keys_file: str = cmd.cli.run("cat ~/.ssh/authorized_keys", hide=True).stdout.strip()
    return keys_file.split("\n")


def ssh_authorized_keys_ensure(*ssh_keys: str) -> List[bool]:
    """
    Добавить несколько ssh-ключей в авторизованные

    :param ssh_keys: ssh-ключи
    :return: Список `True` если ключ был добавлен, `False` если ключ уже был в файле
    """
    return [ssh_authorized_keys_add(x) for x in ssh_keys]


def ssh_copy_id(pubkey_file: str = "~/.ssh/id_rsa.pub") -> bool:
    """
    Добавить публичный ssh-ключ текущего пользователя в авторизованные

    :param pubkey_file: путь до файла с публичным ключем
    :return: `True` если ключ был добавлен, `False` если ключ уже был в файле
    """
    return ssh_authorized_keys_add(open(os.path.expanduser(pubkey_file)).read().strip())


def get_current_user_name() -> str:
    """
    Получить имя текущего пользователя
    """
    id_res: str = cmd.cli.run("id -u -n", hide=True).stdout
    return id_res.strip()


def get_current_user_id() -> int:
    """
    Получить id текущего пользователя
    """
    return int(cmd.cli.run("id -u", hide=True).stdout.strip())


def is_current_user_root() -> bool:
    """
    Проверить что текущий пользователь - `root`
    """
    return get_current_user_id() == 0



from typing import Any

from carnival import global_context
from invoke import Result  # type: ignore


def _run_command(command: str, **kwargs: Any) -> Result:
    assert global_context.conn is not None, "No connection"
    return global_context.conn.run(command, **kwargs)


def run(command: str, **kwargs: Any) -> Result:
    """
    Запустить комманду
    """
    return _run_command(command, **kwargs)


def pty(command: str, **kwargs: Any) -> Result:
    """
    Запустить комманду, используя псевдотерминальную сессию

    См <https://docs.pyinvoke.org/en/latest/api/runners.html>
    """
    return run(command, pty=True, **kwargs)



from io import BytesIO
from typing import Any, Iterable

from carnival import global_context
from carnival.templates import render
from fabric.transfer import Result, Transfer  # type:ignore
from patchwork import transfers  # type:ignore


def rsync(
    source: str, target: str,
    exclude: Iterable[str] = (),
    delete: bool = False, strict_host_keys: bool = True,
    rsync_opts: str = "--progress -pthrvz",
    ssh_opts: str = ''
) -> Result:
    """
    <https://fabric-patchwork.readthedocs.io/en/latest/api/transfers.html#patchwork.transfers.rsync>
    """
    return transfers.rsync(
        c=global_context.conn,
        source=source,
        target=target,
        exclude=exclude,
        delete=delete,
        strict_host_keys=strict_host_keys,
        rsync_opts=rsync_opts,
        ssh_opts=ssh_opts,
    )


def get(remote: str, local: str, preserve_mode: bool = True) -> Result:
    """
    Скачать файл с сервера
    <http://docs.fabfile.org/en/2.5/api/transfer.html#fabric.transfer.Transfer.get>

    :param remote: путь до файла на сервере
    :param local: путь куда сохранить файл
    :param preserve_mode: сохранить права
    """
    t = Transfer(global_context.conn)
    return t.get(remote=remote, local=local, preserve_mode=preserve_mode)


def put(local: str, remote: str, preserve_mode: bool = True) -> Result:
    """
    Закачать файл на сервер
    <http://docs.fabfile.org/en/2.5/api/transfer.html#fabric.transfer.Transfer.put>

    :param local: путь до локального файла
    :param remote: путь куда сохранить на сервере
    :param preserve_mode: сохранить права
    """
    t = Transfer(global_context.conn)
    return t.put(local=local, remote=remote, preserve_mode=preserve_mode)


def put_template(template_path: str, remote: str, **context: Any) -> Result:
    """
    Отрендерить файл с помощью jinja-шаблонов и закачать на сервер
    См раздел templates.

    <http://docs.fabfile.org/en/2.5/api/transfer.html#fabric.transfer.Transfer.put>

    :param template_path: путь до локального файла jinja
    :param remote: путь куда сохранить на сервере
    :param context: контекс для рендеринга jinja2
    """
    filestr = render(template_path=template_path, **context)
    t = Transfer(global_context.conn)
    return t.put(local=BytesIO(filestr.encode()), remote=remote, preserve_mode=False)



from typing import Any, Optional, Type, Union

from fabric import Connection  # type:ignore
from invoke import Context  # type:ignore

from carnival.host import AnyHost

# noinspection PyTypeChecker
conn: Union[Connection, Context, None] = None
# noinspection PyTypeChecker
host: Optional[AnyHost] = None


class SetContext:
    def __init__(self, h: AnyHost):
        self.host = h

    def __enter__(self) -> None:
        global conn
        global host

        assert host is None, f"Cannot set context, while other context active: {host}"
        assert conn is None, f"Cannot set context, while other context active: {conn}"

        conn = self.host.connect()
        host = self.host

    def __exit__(self, exc_type: Type[Any], exc_val: Any, exc_tb: Any) -> None:
        global conn
        global host
        conn = None
        host = None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['carnival', 'carnival.cmd']

package_data = \
{'': ['*']}

install_requires = \
['Click==8.0.3',
 'Fabric==2.6.0',
 'Jinja2==3.0.3',
 'invoke==1.6.0',
 'patchwork==1.0.1',
 'python-dotenv==0.19.2']

entry_points = \
{'console_scripts': ['carnival = carnival.cli:main']}

setup_kwargs = {
    'name': 'carnival',
    'version': '1.3.6',
    'description': 'Fabric-based software provisioning tool',
    'long_description': None,
    'author': 'Dmirty Simonov',
    'author_email': 'demalf@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import os
import sys
from typing import Any, Dict, Iterable, Type

import click
import dotenv

from carnival.task import Task
from carnival.tasks_loader import get_tasks

# Load dotenv first
carnival_dotenv = os.getenv("CARNIVAL_DOTENV", '.env')
try:
    dotenv.load_dotenv(dotenv_path=carnival_dotenv)
except OSError:
    # dotenv file not found
    pass

carnival_tasks_module = os.getenv("CARNIVAL_TASKS_MODULE", "carnival_tasks")


def is_completion_script(complete_var: str) -> bool:
    return os.getenv(complete_var, None) is not None


task_types: Dict[str, Type[Task]] = {}


def except_hook(type: Type[Any], value: Any, traceback: Any) -> None:
    print(f"{type} was raised. You can use --debug flag to see full traceback.")


def main() -> int:
    """
        >>> $ poetry run python -m carnival --help
        >>> Usage: python -m carnival [OPTIONS] {help|test}...
        >>> Options:
        >>> -d, --dry_run  Simulate run
        >>> --debug        Turn on debug mode
        >>> --help         Show this message and exit.
    """
    global task_types

    complete_var = os.getenv("COMPLETE_VAR", "_CARNIVAL_COMPLETE")
    task_types = get_tasks(
        carnival_tasks_module=carnival_tasks_module,
        for_completion=is_completion_script(complete_var)
    )

    @click.command()
    @click.option('-d', '--dry_run', is_flag=True, default=False, help="Simulate run")
    @click.option('--debug', is_flag=True, default=False, help="Turn on debug mode")
    @click.argument('tasks', required=True, type=click.Choice(list(task_types.keys())), nargs=-1)
    def cli(dry_run: bool, debug: bool, tasks: Iterable[str]) -> None:
        if debug is True:
            print("Debug mode on.")
        else:
            sys.excepthook = except_hook

        for task in tasks:
            task_types[task](dry_run=dry_run).run()

    cli(complete_var=complete_var)
    return 0



import abc
from typing import Any, no_type_check

from carnival.context import build_context, build_kwargs
from carnival.host import AnyHost


class Step:
    """
    Объект, предназначенный для выполнения группы комманд с какой-то целью.
    Вызывается из класса `carnival.Task` для выполнения комманд (`carnival.cmd`) на определенных хостах.

    Может требовать наличие определенных контекстных переменных для работы, указав их в аргументах метода `run`.
    Может вернуть значение для дальнейшего использования.

    В следующем примере переменная `disk_name` будет передана в run, а `install` пропущена.

    >>> host = Host(
    >>>     #  Адрес
    >>>     "1.2.3.4",
    >>>
    >>>     # Контекст хоста
    >>>     disk_name="/dev/sda1", install=['nginx', 'htop', ]
    >>> )
    >>> ...
    >>> class DiskUsage(Step):
    >>>     def run(self, disk_name: str):
    >>>         ...

    """
    def __init__(self, **context: Any):
        """
        :param context: Переменные контекста, назначенные при вызове Шага
        """
        self.context = context

    def run_with_context(self, host: AnyHost) -> Any:
        context = build_context(self, host)
        kwargs = build_kwargs(self.run, context)
        return self.run(**kwargs)  # type: ignore

    @abc.abstractmethod
    @no_type_check
    def run(self, **kwargs) -> None:
        """
        Метод который нужно определить для выполнения комманд

        :param kwargs: Автоматические подставляемые переменные контекста, поддерживается `**kwargs`
        """

        raise NotImplementedError



import inspect
import os
from itertools import chain
from typing import TYPE_CHECKING, Any, Callable, Dict, List

if TYPE_CHECKING:
    from carnival.host import AnyHost
    from carnival.step import Step


class PassAllArgs(BaseException):
    pass


def _get_arg_names(fn: Callable[..., Any]) -> List[str]:
    arg_names: List[str] = []
    spec = inspect.getfullargspec(fn)

    if spec.varargs is not None:
        raise ValueError("*args is not supported for autocontext")

    if spec.varkw is not None:
        # Not actual if kwargs argument exists
        raise PassAllArgs()

    arg_names += spec.args
    arg_names += spec.kwonlyargs

    if 'self' in arg_names:
        arg_names.remove('self')

    return arg_names


def build_kwargs(fn: Callable[..., Any], context: Dict[str, Any]) -> Dict[str, Any]:
    try:
        arg_names: List[str] = _get_arg_names(fn)
    except PassAllArgs:
        # Pass all context if kwargs var exists
        return context.copy()

    kwargs = {}
    for context_name, context_val in context.items():
        if context_name in arg_names:
            kwargs[context_name] = context_val
    return kwargs


def build_context(step: 'Step', host: 'AnyHost') -> Dict[str, Any]:
    run_context: Dict[str, Any] = {'host': host}

    env_prefix = "CARNIVAL_CTX_"
    # Build context from environment variables
    for env_name, env_val in os.environ.items():
        if env_name.startswith(env_prefix):
            run_context[env_name[len(env_prefix):]] = env_val

    for var_name, var_val in chain(host.context.items(), step.context.items()):
        if isinstance(var_val, context_ref):
            try:
                run_context[var_name] = run_context[var_val.context_var_name]
            except KeyError as e:
                raise KeyError(f"There is no '{var_val.context_var_name}' variable in context") from e
        else:
            run_context[var_name] = var_val

    return run_context


class context_ref:
    """
    Ссылка на другую переменную контекста

    Например, так можно передать в переменную c именем `name` `Step` другую переменную контекста с именем `dns_domain`
    >>> TestStep(
    >>>   name=context_ref('dns_domain'),
    >>> ),

    """
    def __init__(self, context_var_name: str):
        self.context_var_name = context_var_name



from carnival import cmd
from invoke import Result  # type: ignore


def daemon_reload() -> Result:
    """
    Перегрузить systemd
    """
    return cmd.cli.run("sudo systemctl --system daemon-reload")


def start(service_name: str, reload_daemon: bool = False) -> Result:
    """
    Запустить сервис

    :param service_name: имя сервиса
    :param reload_daemon: перегрузить systemd
    """

    if reload_daemon:
        daemon_reload()
    return cmd.cli.run(f"sudo systemctl start {service_name}")


def stop(service_name: str, reload_daemon: bool = False) -> Result:
    """
    Остановить сервис

    :param service_name: имя сервиса
    :param reload_daemon: перегрузить systemd
    """
    if reload_daemon:
        daemon_reload()
    return cmd.cli.run(f"sudo systemctl stop {service_name}")


def restart(service_name: str) -> Result:
    """
    Перезапустить сервис

    :param service_name: имя сервиса
    """
    return cmd.cli.run(f"sudo systemctl restart {service_name}")


def enable(service_name: str, reload_daemon: bool = False, start_now: bool = True) -> Result:
    """
    Добавить сервис в автозапуск

    :param service_name: имя сервиса
    :param reload_daemon: перегрузить systemd
    :param start_now: запустить сервис после добавления
    """
    if reload_daemon:
        daemon_reload()

    res = cmd.cli.run(f"sudo systemctl enable {service_name}")

    if start_now:
        start(service_name)

    return res


def disable(service_name: str, reload_daemon: bool = False, stop_now: bool = True) -> Result:
    """
    Убрать сервис из автозапуска

    :param service_name: имя сервиса
    :param reload_daemon: перегрузить systemd
    :param stop_now: Остановить сервис
    """

    if reload_daemon:
        daemon_reload()

    res = cmd.cli.run(f"sudo systemctl disable {service_name}")

    if stop_now:
        stop(service_name)

    return res



"""
Модуль carnival.cmd содержит базовые комманды для взаимодействия с сервером.
Его цель - оставаться простым и помогать в написании шагов (Step).

Для написания сложных сценариев предполагается использовать шаги(Step).

Основные шаги доступны в отдельном репозитории: <https://github.com/carnival-org/carnival-contrib>.
"""

from carnival.cmd import cli, system, systemd, apt, transfer, fs


__all__ = [
    'cli',
    'system',
    'systemd',
    'apt',
    'transfer',
    'fs',
]
